"""This module defines multiple implementations of the abstract class :class:`~dowhy.gcm.graph.FunctionalCausalModel`
(FCM)

Classes in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

import copy
from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np

from dowhy.gcm.graph import FunctionalCausalModel, InvertibleFunctionalCausalModel, StochasticModel
from dowhy.gcm.util.general import is_categorical, shape_into_2d


class PredictionModel:
    """Represents general prediction model implementations. Each prediction model should provide a fit and a predict
    method."""

    @abstractmethod
    def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
        raise NotImplementedError

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def clone(self):
        """
        Clones the prediction model using the same hyper parameters but not fitted.

        :return: An unfitted clone of the prediction model.
        """
        raise NotImplementedError


class ClassificationModel(PredictionModel):
    @abstractmethod
    def predict_probabilities(self, X: np.array) -> np.ndarray:
        raise NotImplementedError

    @property
    @abstractmethod
    def classes(self) -> List[str]:
        raise NotImplementedError


class InvertibleFunction:
    @abstractmethod
    def evaluate(self, X: np.ndarray) -> np.ndarray:
        """Applies the function on the input."""
        raise NotImplementedError

    @abstractmethod
    def evaluate_inverse(self, X: np.ndarray) -> np.ndarray:
        """Returns the outcome of applying the inverse of the function on the inputs."""
        raise NotImplementedError


class PostNonlinearModel(InvertibleFunctionalCausalModel):
    """
    Represents an post nonlinear FCM, i.e. models of the form:
        Y = g(f(X) + N),
    where X are parent nodes of the target node Y, f an arbitrary prediction model expecting inputs from the
    parents X, N a noise variable and g an invertible function.
    """

    def __init__(
        self, prediction_model: PredictionModel, noise_model: StochasticModel, invertible_function: InvertibleFunction
    ) -> None:
        """
        :param prediction_model: The prediction model f.
        :param invertible_function: The invertible function g.
        :param noise_model: The StochasticModel to describe the distribution of the noise N.
        """
        self._prediction_model = prediction_model
        self._noise_model = noise_model
        self._invertible_function = invertible_function

    def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
        """Fits the post non-linear model of the form Y = g(f(X) + N). Here, this consists of three steps given
        samples from (X, Y):

            1. Transform Y via the inverse of g: g^-1(Y) = f(X) + N
            2. Fit the model for f on (X, g^-1(Y))
            3. Reconstruct N based on the residual N = g^-1(Y) - f(X)

        Note that the noise here can be inferred uniquely if the model assumption holds.

        :param X: Samples from the input X.
        :param Y: Samples from the target Y.
        :return: None
        """
        X, Y = shape_into_2d(X, Y)

        self._prediction_model.fit(X=X, Y=self._invertible_function.evaluate_inverse(Y))
        self._noise_model.fit(X=self.estimate_noise(Y, X))

    def estimate_noise(self, target_samples: np.ndarray, parent_samples: np.ndarray) -> np.ndarray:
        """Reconstruct the noise given samples from (X, Y). This is done by:

            1. Transform Y via the inverse of g: g^-1(Y) = f(X) + N
            2. Return the residual g^-1(Y) - f(X)

        :param target_samples: Samples from the input X.
        :param parent_samples: Samples from the target Y.
        :return: The reconstructed noise based on the given samples.
        """
        target_samples, parent_samples = shape_into_2d(target_samples, parent_samples)

        return self._invertible_function.evaluate_inverse(target_samples) - self._prediction_model.predict(
            parent_samples
        )

    def draw_noise_samples(self, num_samples: int) -> np.ndarray:
        """Draws samples from the noise distribution N.

        :param num_samples: Number of noise samples.
        :return: A numpy array containing num_samples samples from the noise.
        """
        return self._noise_model.draw_samples(num_samples)

    def evaluate(self, parent_samples: np.ndarray, noise_samples: np.ndarray) -> np.ndarray:
        """Evaluates the post non-linear model given samples (X, N). This is done by:

            1. Evaluate f(X)
            2. Evaluate f(X) + N
            3. Return g(f(X) + N)

        :param parent_samples: Samples from the inputs X.
        :param noise_samples: Samples from the noise N.
        :return: The Y values based on the given samples.
        """
        parent_samples, noise_samples = shape_into_2d(parent_samples, noise_samples)
        predictions = shape_into_2d(self._prediction_model.predict(parent_samples))

        return self._invertible_function.evaluate(predictions + noise_samples)

    def __str__(self) -> str:
        return "%s with %s and an %s" % (
            self.__class__.__name__,
            self._prediction_model.__class__.__name__,
            self._invertible_function.__class__.__name__,
        )

    def clone(self):
        return PostNonlinearModel(
            prediction_model=self._prediction_model.clone(),
            noise_model=self._noise_model.clone(),
            invertible_function=copy.deepcopy(self._invertible_function),
        )

    @property
    def prediction_model(self) -> PredictionModel:
        return self._prediction_model

    @property
    def noise_model(self) -> StochasticModel:
        return self._noise_model

    @property
    def invertible_function(self) -> InvertibleFunction:
        return self._invertible_function


class AdditiveNoiseModel(PostNonlinearModel):
    """Represents the continuous functional causal model of the form
        Y = f(X) + N,
    where X is the input (typically, direct causal parents of Y) and the noise N is assumed to be independent of X. This
    is a special instance of a :py:class:`PostNonlinearModel <dowhy.gcm.PostNonlinearModel>` where the function g is the
    identity function.

    Given joint samples from (X, Y), this model can be fitted by first training a model f (e.g. using least squares
    regression) and then reconstruct N by N = Y - f(X), i.e. using the residual.
    """

    def __init__(self, prediction_model: PredictionModel, noise_model: Optional[StochasticModel] = None) -> None:
        if noise_model is None:
            from dowhy.gcm.stochastic_models import EmpiricalDistribution

            noise_model = EmpiricalDistribution()

        from dowhy.gcm.ml.regression import InvertibleIdentityFunction

        super(AdditiveNoiseModel, self).__init__(
            prediction_model=prediction_model, noise_model=noise_model, invertible_function=InvertibleIdentityFunction()
        )

    def clone(self):
        return AdditiveNoiseModel(prediction_model=self.prediction_model.clone(), noise_model=self.noise_model.clone())


class ProbabilityEstimatorModel(ABC):
    @abstractmethod
    def estimate_probabilities(self, parent_samples: np.ndarray) -> np.ndarray:
        raise NotImplementedError


class ClassifierFCM(FunctionalCausalModel, ProbabilityEstimatorModel):
    """Represents the categorical functional causal model of the form
        Y = f(X, N),
    where X is the input (typically, direct causal parents of Y) and the noise N here is uniform on [0, 1]. The model
    is mostly based on a standard classification model that outputs probabilities. In order to generate a new random
    sample given an input x, the return value y is uniformly sampled based on the class probabilities p(y | x). Here,
    the noise is used to make this sampling process deterministic by using the cumulative distribution functions defined
    by the given inputs.
    """

    def __init__(self, classifier_model: Optional[ClassificationModel] = None) -> None:
        self._classifier_model = classifier_model

        if classifier_model is None:
            from dowhy.gcm.ml.classification import create_hist_gradient_boost_classifier

            self._classifier_model = create_hist_gradient_boost_classifier()

    def draw_noise_samples(self, num_samples: int) -> np.ndarray:
        """Returns uniformly sampled values on [0, 1].

        :param num_samples: Number of noise samples.
        :return: Noise samples on [0, 1].
        """
        return shape_into_2d(np.random.uniform(0, 1, num_samples))

    def evaluate(self, parent_samples: np.ndarray, noise_samples: np.ndarray) -> np.ndarray:
        """Evaluates the model Y = f(X, N), where X are the parent_samples and N the noise_samples. Here, the
        cumulative distribution functions are defined by the parent_samples. For instance, lets say we have 2
        classes, n = 0.7 and an input x with p(y = 0| x) = 0.6 and p(y = 1| x) = 0.4, then we get y = 1 as a return
        value. This is because p(y = 0| x) < n <= 1.0, i.e. n falls into the bucket that is spanned by p(y = 1| x).

        :param parent_samples: Samples from the inputs X.
        :param noise_samples: Samples from the noise on [0, 1].
        :return: Class labels Y based on the inputs and noise.
        """
        noise_samples = shape_into_2d(noise_samples)
        probabilities = self.estimate_probabilities(parent_samples)

        probabilities = np.cumsum(probabilities, axis=1) - noise_samples
        probabilities[probabilities < 0] = 1

        return shape_into_2d(np.array(self.get_class_names(np.argmin(probabilities, axis=1))))

    def estimate_probabilities(self, parent_samples: np.ndarray) -> np.ndarray:
        """Returns the class probabilities for the given parent_samples.

        :param parent_samples: Samples from inputs X.
        :return: A nxd numpy matrix with class probabilities for each sample, where n is the number of samples and d
                 the number of classes. Here, array entry A[i][j] corresponds to the i-th sample indicating the
                 probability of the j-th class.
        """
        return self._classifier_model.predict_probabilities(parent_samples)

    def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
        """Fits the underlying classification model.

        :param X: Input samples.
        :param Y: Target labels.
        :return: None
        """
        X, Y = shape_into_2d(X, Y)

        if not is_categorical(Y):
            raise ValueError("The target data needs to be categorical in the form of strings!")

        self._classifier_model.fit(X=X, Y=Y)

    def clone(self):
        return ClassifierFCM(classifier_model=self._classifier_model.clone())

    def get_class_names(self, class_indices: np.ndarray) -> List[str]:
        return [self._classifier_model.classes[index] for index in class_indices]

    @property
    def classifier_model(self) -> ClassificationModel:
        return self._classifier_model



from pytest import approx

from dowhy.gcm.util.pygraphviz import _calc_arrow_width


def test_calc_arrow_width():
    assert _calc_arrow_width(0.4, max_strength=0.5) == approx(3.3, abs=0.01)
    assert _calc_arrow_width(0.2, max_strength=0.5) == approx(1.7, abs=0.01)
    assert _calc_arrow_width(-0.2, max_strength=0.5) == approx(1.7, abs=0.01)
    assert _calc_arrow_width(0.5, max_strength=0.5) == approx(4.1, abs=0.01)
    assert _calc_arrow_width(0.35, max_strength=0.5) == approx(2.9, abs=0.01)
    assert _calc_arrow_width(100, max_strength=101) == approx(4.06, abs=0.01)



import string
from importlib import import_module

from dowhy.causal_refuter import CausalRefuter
from dowhy.causal_refuters.add_unobserved_common_cause import (
    AddUnobservedCommonCause,
    sensitivity_e_value,
    sensitivity_linear_partial_r2,
    sensitivity_non_parametric_partial_r2,
    sensitivity_simulation,
)
from dowhy.causal_refuters.bootstrap_refuter import BootstrapRefuter, refute_bootstrap
from dowhy.causal_refuters.data_subset_refuter import DataSubsetRefuter, refute_data_subset
from dowhy.causal_refuters.dummy_outcome_refuter import DummyOutcomeRefuter, refute_dummy_outcome
from dowhy.causal_refuters.placebo_treatment_refuter import PlaceboTreatmentRefuter, refute_placebo_treatment
from dowhy.causal_refuters.random_common_cause import RandomCommonCause, refute_random_common_cause
from dowhy.causal_refuters.refute_estimate import refute_estimate


def get_class_object(method_name, *args, **kwargs):
    # from https://www.bnmetrics.com/blog/factory-pattern-in-python3-simple-version
    try:
        module_name = method_name
        class_name = string.capwords(method_name, "_").replace("_", "")

        refuter_module = import_module("." + module_name, package="dowhy.causal_refuters")
        refuter_class = getattr(refuter_module, class_name)
        assert issubclass(refuter_class, CausalRefuter)

    except (AttributeError, AssertionError, ImportError):
        raise ImportError("{} is not an existing causal refuter.".format(method_name))
    return refuter_class


__all__ = [
    "AddUnobservedCommonCause",
    "BootstrapRefuter",
    "DataSubsetRefuter",
    "PlaceboTreatmentRefuter",
    "RandomCommonCause",
    "DummyOutcomeRefuter",
    "refute_bootstrap",
    "refute_data_subset",
    "refute_random_common_cause",
    "refute_placebo_treatment",
    "sensitivity_simulation",
    "sensitivity_linear_partial_r2",
    "sensitivity_non_parametric_partial_r2",
    "sensitivity_e_value",
    "refute_dummy_outcome",
    "refute_estimate",
]



from typing import Any, Callable, List, Optional, Tuple

import networkx as nx
import numpy as np
import pandas as pd

from dowhy.gcm.cms import InvertibleStructuralCausalModel, ProbabilisticCausalModel, StructuralCausalModel
from dowhy.gcm.fcms import PredictionModel
from dowhy.gcm.graph import get_ordered_predecessors, is_root_node, node_connected_subgraph_view, validate_causal_dag
from dowhy.gcm.util.general import shape_into_2d


def compute_data_from_noise(causal_model: StructuralCausalModel, noise_data: pd.DataFrame) -> pd.DataFrame:
    validate_causal_dag(causal_model.graph)

    sorted_nodes = list(nx.topological_sort(causal_model.graph))
    data = pd.DataFrame(np.empty((noise_data.shape[0], len(sorted_nodes))), columns=sorted_nodes)

    for node in sorted_nodes:
        if is_root_node(causal_model.graph, node):
            data[node] = noise_data[node].to_numpy()
        else:
            data[node] = causal_model.causal_mechanism(node).evaluate(
                data[get_ordered_predecessors(causal_model.graph, node)].to_numpy(), noise_data[node].to_numpy()
            )

    return data


def compute_noise_from_data(causal_model: InvertibleStructuralCausalModel, observed_data: pd.DataFrame) -> pd.DataFrame:
    validate_causal_dag(causal_model.graph)

    sorted_noise = list(nx.topological_sort(causal_model.graph))
    noise = pd.DataFrame(np.empty((observed_data.shape[0], len(sorted_noise))), columns=sorted_noise)

    for node in sorted_noise:
        if is_root_node(causal_model.graph, node):
            noise[node] = observed_data[node].to_numpy()
        else:
            noise[node] = causal_model.causal_mechanism(node).estimate_noise(
                observed_data[node].to_numpy(), _parent_samples_of(node, causal_model, observed_data)
            )

    return noise


def get_noise_dependent_function(
    causal_model: StructuralCausalModel,
    target_node: Any,
    approx_prediction_model: Optional[PredictionModel] = None,
    num_training_samples: int = 20000,
) -> Tuple[Callable[[np.ndarray], np.ndarray], List[Any]]:
    """Returns a function that represents the given target_node and that is only dependent on upstream noise nodes.
    This is, Y = f(N_0, N_1, ..., N_n), where Y is the target node and N_i the noise node of an upstream node. Since
    the order of the noise variables can be ambiguous, this method also returns a list with the expected order of the
    noise variables indicated by the name of the corresponding node. For instance:
    Lets say we have a target variable X4 which has only 2 other upstream variables X1 and X3, then this methods returns
    a callable that represents X4 = f(N_1, N_3, N_4). This callable expects a numpy array as input. Since the order
    of the columns in this array is unclear, the method also returns a list with the names of the corresponding columns,
    e.g. ['X1', 'X3', 'X4']. Note that the noise of X4 will also be an input variable here.

    If an approx_prediction_model is given, the model is used instead to train a model from scratch to represent f.
    Typically, it would be much faster to evaluate this model than propagating through the graph. The model is trained
    based on generated noise and target samples from the given causal graph. In theory, the approximated version of f
    should be close to the "true" one based on the underlying SCMs, but there can still be (significant) differences,
    especially if the provided model is inappropriate for representing f.

    Note: All nodes in the graph that have no direct path to the target node are omitted. The noise node of the
          target variable itself will also be included here.

    :param causal_model: A structural causal model.
    :param target_node: The target node for which the function f should be returned.
    :param approx_prediction_model: Prediction model for approximating f. The model is trained based on drawn noise and
                                    target samples.
    :param num_training_samples: Number of drawn samples for training the predictor based on the provided
                                 approx_prediction_model.
                                 Note: This parameter is ignored if approx_prediction_model is None.
    :return: A tuple, where the first value is a callable (the function f) that expects a numpy array X as input and
    the second value is a list with nodes that represents the expected order of the columns in X.
    """
    validate_causal_dag(causal_model.graph)

    if approx_prediction_model is not None:
        return _get_approx_noise_dependent_function(
            StructuralCausalModel(node_connected_subgraph_view(causal_model.graph, target_node)),
            target_node,
            approx_prediction_model,
            num_training_samples,
        )
    else:
        return _get_exact_noise_dependent_function(
            StructuralCausalModel(node_connected_subgraph_view(causal_model.graph, target_node)), target_node
        )


def _get_exact_noise_dependent_function(
    causal_model: StructuralCausalModel, target_node: Any
) -> Tuple[Callable[[np.ndarray], np.ndarray], List[Any]]:
    nodes_order = list(nx.topological_sort(causal_model.graph))

    def predict_method(noise_samples: np.ndarray) -> np.ndarray:
        return compute_data_from_noise(causal_model, pd.DataFrame(noise_samples, columns=[x for x in nodes_order]))[
            target_node
        ].to_numpy()

    return predict_method, nodes_order


def _get_approx_noise_dependent_function(
    causal_model: StructuralCausalModel,
    target_node: Any,
    approx_prediction_model: PredictionModel,
    num_training_samples: int,
) -> Tuple[Callable[[np.ndarray], np.ndarray], List[Any]]:
    nodes_order = list(nx.topological_sort(causal_model.graph))

    node_samples, noise_samples = noise_samples_of_ancestors(causal_model, target_node, num_training_samples)

    approx_prediction_model.fit(
        shape_into_2d(noise_samples[nodes_order].to_numpy()), shape_into_2d(node_samples[target_node].to_numpy())
    )

    return approx_prediction_model.predict, nodes_order


def noise_samples_of_ancestors(
    causal_model: StructuralCausalModel, target_node: Any, num_samples: int
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    sorted_nodes = list(nx.topological_sort(causal_model.graph))
    all_ancestors_of_node = nx.ancestors(causal_model.graph, target_node)
    all_ancestors_of_node.update({target_node})

    drawn_samples = pd.DataFrame(np.empty((num_samples, len(sorted_nodes))), columns=sorted_nodes)
    drawn_noise_samples = pd.DataFrame(np.empty((num_samples, len(sorted_nodes))), columns=sorted_nodes)

    for node in sorted_nodes:
        if node not in all_ancestors_of_node:
            continue

        if is_root_node(causal_model.graph, node):
            noise = causal_model.causal_mechanism(node).draw_samples(num_samples)
            drawn_noise_samples[node] = noise
            drawn_samples[node] = noise
        else:
            noise = causal_model.causal_mechanism(node).draw_noise_samples(num_samples)
            drawn_noise_samples[node] = noise
            drawn_samples[node] = causal_model.causal_mechanism(node).evaluate(
                _parent_samples_of(node, causal_model, drawn_samples), noise
            )

        if node == target_node:
            break

    return drawn_samples, drawn_noise_samples


def _parent_samples_of(node: Any, scm: ProbabilisticCausalModel, samples: pd.DataFrame) -> np.ndarray:
    return samples[get_ordered_predecessors(scm.graph, node)].to_numpy()



import random

import numpy as np
import pytest
from flaky import flaky

from dowhy.gcm.independence_test import regression_based
from dowhy.gcm.util.general import set_random_seed
from tests.gcm.independence_test.test_kernel import _generate_categorical_data


@pytest.fixture
def preserve_random_generator_state():
    numpy_state = np.random.get_state()
    random_state = random.getstate()
    yield
    np.random.set_state(numpy_state)
    random.setstate(random_state)


@flaky(max_runs=5)
def test_regression_based_conditional_independence_test_independent():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert regression_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_regression_based_conditional_independence_test_dependent():
    z = np.random.randn(1000, 1)
    w = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert 0.05 > regression_based(x, y, w)


@flaky(max_runs=5)
def test_regression_based_conditional_independence_test_categorical_independent():
    x, y, z = _generate_categorical_data()

    assert regression_based(x, y, z) > 0.05


@flaky(max_runs=5)
def test_regression_based_conditional_independence_test_categorical_dependent():
    x, y, z = _generate_categorical_data()

    assert regression_based(x, z, y) < 0.05


@flaky(max_runs=5)
def test_regression_based_pairwise_independence_test_independent():
    z = np.random.randn(1000, 1)
    w = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))

    assert regression_based(x, w) > 0.05


@flaky(max_runs=5)
def test_regression_based_pairwise_independence_test_dependent():
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    assert regression_based(x, y) < 0.05


@flaky(max_runs=5)
def test_regression_based_pairwise_independence_test_categorical_independent():
    x = np.random.normal(0, 1, 1000)
    y = np.random.choice(2, 1000).astype(str)

    assert regression_based(x, y) > 0.05


@flaky(max_runs=5)
def test_regression_based_pairwise_independence_test_categorical_dependent():
    x = np.random.normal(0, 1, 1000)
    y = []

    for v in x:
        if v > 0:
            y.append(0)
        else:
            y.append(1)
    y = np.array(y).astype(str)

    assert regression_based(x, y, num_components_all_inputs=10) < 0.05


def test_regression_based_conditional_independence_parallelization(preserve_random_generator_state):
    z = np.random.randn(1000, 1)
    x = np.exp(z + np.random.rand(1000, 1))
    y = np.exp(z + np.random.rand(1000, 1))

    set_random_seed(0)
    p_value_1 = regression_based(x, y, z)
    set_random_seed(0)
    p_value_2 = regression_based(x, y, z)

    assert p_value_1 == p_value_2


@flaky(max_runs=5)
def test_regression_based_conditional_independence_test_categorical_dependent():
    x, y, z = _generate_categorical_data()

    assert regression_based(x, z, y) < 0.05



import copy
import logging
from enum import Enum
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm.auto import tqdm

from dowhy.causal_estimator import CausalEstimate
from dowhy.causal_estimators.econml import Econml
from dowhy.causal_estimators.instrumental_variable_estimator import InstrumentalVariableEstimator
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.causal_refuter import CausalRefutation, CausalRefuter, test_significance
from dowhy.utils.api import parse_state

logger = logging.getLogger(__name__)


# Default value of the p value taken for the distribution
DEFAULT_PROBABILITY_OF_BINOMIAL = 0.5
# Number of Trials: Number of cointosses to understand if a sample gets the treatment
DEFAULT_NUMBER_OF_TRIALS = 1
# Mean of the Normal Distribution
DEFAULT_MEAN_OF_NORMAL = 0
# Standard Deviation of the Normal Distribution
DEFAULT_STD_DEV_OF_NORMAL = 0


class PlaceboType(Enum):
    DEFAULT = "Random Data"
    PERMUTE = "permute"


class PlaceboTreatmentRefuter(CausalRefuter):
    """Refute an estimate by replacing treatment with a randomly-generated placebo variable.

    Supports additional parameters that can be specified in the refute_estimate() method. For joblib-related parameters (n_jobs, verbose), please refer to the joblib documentation for more details (https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html).

    :param placebo_type: Default is to generate random values for the treatment. If placebo_type is "permute", then the original treatment values are permuted by row.
    :type placebo_type: str, optional

    :param num_simulations: The number of simulations to be run, which is ``CausalRefuter.DEFAULT_NUM_SIMULATIONS`` by default
    :type num_simulations: int, optional

    :param random_state: The seed value to be added if we wish to repeat the same random behavior. If we want to repeat the same behavior we push the same seed in the psuedo-random generator.
    :type random_state: int, RandomState, optional

    :param n_jobs: The maximum number of concurrently running jobs. If -1 all CPUs are used. If 1 is given, no parallel computing code is used at all (this is the default).
    :type n_jobs: int, optional

    :param verbose: The verbosity level: if non zero, progress messages are printed. Above 50, the output is sent to stdout. The frequency of the messages increases with the verbosity level. If it more than 10, all iterations are reported. The default is 0.
    :type verbose: int, optional
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._placebo_type = kwargs.pop("placebo_type", None)
        if self._placebo_type is None:
            self._placebo_type = "Random Data"
        self._num_simulations = kwargs.pop("num_simulations", CausalRefuter.DEFAULT_NUM_SIMULATIONS)
        self._random_state = kwargs.pop("random_state", None)

        self.logger = logging.getLogger(__name__)

    def refute_estimate(self, show_progress_bar=False):
        refute = refute_placebo_treatment(
            data=self._data,
            target_estimand=self._target_estimand,
            estimate=self._estimate,
            treatment_names=self._treatment_name,
            num_simulations=self._num_simulations,
            placebo_type=PlaceboType(self._placebo_type),
            random_state=self._random_state,
            show_progress_bar=show_progress_bar,
            n_jobs=self._n_jobs,
            verbose=self._verbose,
        )

        refute.add_refuter(self)

        return refute


def _refute_once(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    treatment_names: List[str],
    type_dict: Dict,
    placebo_type: PlaceboType = PlaceboType.DEFAULT,
    random_state: Optional[np.random.RandomState] = None,
):
    if placebo_type == PlaceboType.PERMUTE:
        permuted_idx = None
        if random_state is None:
            permuted_idx = np.random.choice(data.shape[0], size=data.shape[0], replace=False)

        else:
            permuted_idx = random_state.choice(data.shape[0], size=data.shape[0], replace=False)
        new_treatment = data[treatment_names].iloc[permuted_idx].values
        if target_estimand.identifier_method.startswith("iv"):
            new_instruments_values = data[estimate.estimator.estimating_instrument_names].iloc[permuted_idx].values
            new_instruments_df = pd.DataFrame(
                new_instruments_values,
                columns=["placebo_" + s for s in data[estimate.estimator.estimating_instrument_names].columns],
            )
    else:
        if "float" in type_dict[treatment_names[0]].name:
            logger.info(
                "Using a Normal Distribution with Mean:{} and Variance:{}".format(
                    DEFAULT_MEAN_OF_NORMAL,
                    DEFAULT_STD_DEV_OF_NORMAL,
                )
            )
            new_treatment = np.random.randn(data.shape[0]) * DEFAULT_STD_DEV_OF_NORMAL + DEFAULT_MEAN_OF_NORMAL

        elif "bool" in type_dict[treatment_names[0]].name:
            logger.info(
                "Using a Binomial Distribution with {} trials and {} probability of success".format(
                    DEFAULT_NUMBER_OF_TRIALS,
                    DEFAULT_PROBABILITY_OF_BINOMIAL,
                )
            )
            new_treatment = np.random.binomial(
                DEFAULT_NUMBER_OF_TRIALS,
                DEFAULT_PROBABILITY_OF_BINOMIAL,
                data.shape[0],
            ).astype(bool)

        elif "int" in type_dict[treatment_names[0]].name:
            logger.info(
                "Using a Discrete Uniform Distribution lying between {} and {}".format(
                    data[treatment_names[0]].min(), data[treatment_names[0]].max()
                )
            )
            new_treatment = np.random.randint(
                low=data[treatment_names[0]].min(), high=data[treatment_names[0]].max() + 1, size=data.shape[0]
            )

        elif "category" in type_dict[treatment_names[0]].name:
            categories = data[treatment_names[0]].unique()
            logger.info("Using a Discrete Uniform Distribution with the following categories:{}".format(categories))
            sample = np.random.choice(categories, size=data.shape[0])
            new_treatment = pd.Series(sample, index=data.index).astype("category")

    # Create a new column in the data by the name of placebo
    new_data = data.assign(placebo=new_treatment)
    if target_estimand.identifier_method.startswith("iv"):
        new_data = pd.concat((new_data, new_instruments_df), axis=1)
    # Sanity check the data
    logger.debug(new_data[0:10])
    new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
    new_estimator.fit(
        new_data,
        target_estimand.treatment_variable,
        target_estimand.outcome_variable,
        estimate.estimator._effect_modifier_names,
        **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
    )
    new_effect = new_estimator.estimate_effect(
        control_value=estimate.control_value,
        treatment_value=estimate.treatment_value,
        target_units=estimate.estimator._target_units,
    )
    return new_effect.value


def refute_placebo_treatment(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    treatment_names: List,
    num_simulations: int = 100,
    placebo_type: PlaceboType = PlaceboType.DEFAULT,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    show_progress_bar: bool = False,
    n_jobs: int = 1,
    verbose: int = 0,
    **_,
) -> CausalRefutation:
    """Refute an estimate by replacing treatment with a randomly-generated placebo variable.

    :param data: pd.DataFrame: Data to run the refutation
    :param target_estimand: IdentifiedEstimand: Identified estimand to run the refutation
    :param estimate: CausalEstimate: Estimate to run the refutation
    :param treatment_names: list: List of treatments
    :param num_simulations: The number of simulations to be run, which defaults to ``CausalRefuter.DEFAULT_NUM_SIMULATIONS``
    :param placebo_type: Default is to generate random values for the treatment. If placebo_type is "permute", then the original treatment values are permuted by row.
    :param random_state: The seed value to be added if we wish to repeat the same random behavior. If we want to repeat the same behavior we push the same seed in the psuedo-random generator.
    :param n_jobs: The maximum number of concurrently running jobs. If -1 all CPUs are used. If 1 is given, no parallel computing code is used at all (this is the default).
    :param verbose: The verbosity level: if non zero, progress messages are printed. Above 50, the output is sent to stdout. The frequency of the messages increases with the verbosity level. If it more than 10, all iterations are reported. The default is 0.
    """

    if isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)

    # only permute is supported for iv methods
    if target_estimand.identifier_method.startswith("iv"):
        if placebo_type != PlaceboType.PERMUTE:
            logger.error(
                "Only placebo_type=''permute'' is supported for creating placebo for instrumental variable estimation methods"
            )
            raise ValueError(
                "Only placebo_type=''permute'' is supported for creating placebo for instrumental variable estimation methods."
            )

    # For IV methods, the estimating_instrument_names should also be
    # changed. Create a copy to avoid modifying original object
    if isinstance(estimate, InstrumentalVariableEstimator):
        estimate = copy.deepcopy(estimate)
        estimate.iv_instrument_name = ["placebo_" + s for s in parse_state(estimate.iv_instrument_name)]

    # We need to change the identified estimand
    # We make a copy as a safety measure, we don't want to change the
    # original DataFrame
    identified_estimand = copy.deepcopy(target_estimand)
    identified_estimand.treatment_variable = ["placebo"]

    if target_estimand.identifier_method.startswith("iv"):
        identified_estimand.instrumental_variables = [
            "placebo_" + s for s in identified_estimand.instrumental_variables
        ]

    logger.info("Refutation over {} simulated datasets of {} treatment".format(num_simulations, placebo_type))

    type_dict = dict(data.dtypes)

    # Run refutation in parallel
    sample_estimates = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(_refute_once)(
            data, identified_estimand, estimate, treatment_names, type_dict, placebo_type, random_state
        )
        for _ in tqdm(
            range(num_simulations),
            disable=not show_progress_bar,
            colour=CausalRefuter.PROGRESS_BAR_COLOR,
            desc="Refuting Estimates: ",
        )
    )

    sample_estimates = np.array(sample_estimates)

    refute = CausalRefutation(
        estimate.value, np.mean(sample_estimates), refutation_type="Refute: Use a Placebo Treatment"
    )

    # Note: We hardcode the estimate value to ZERO as we want to check if it falls in the distribution of the refuter
    # Ideally we should expect that ZERO should fall in the distribution of the effect estimates as we have severed any causal
    # relationship between the treatment and the outcome.
    dummy_estimator = CausalEstimate(
        estimate=0,
        control_value=estimate.control_value,
        treatment_value=estimate.treatment_value,
        target_estimand=estimate.target_estimand,
        realized_estimand_expr=estimate.realized_estimand_expr,
    )

    refute.add_significance_test_results(test_significance(dummy_estimator, sample_estimates))

    return refute



import warnings
from enum import Enum, auto
from functools import partial
from typing import Callable, List, Optional, Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn import metrics
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.model_selection import KFold, train_test_split
from sklearn.preprocessing import MultiLabelBinarizer

from dowhy.gcm import config
from dowhy.gcm.cms import ProbabilisticCausalModel
from dowhy.gcm.fcms import AdditiveNoiseModel, ClassificationModel, ClassifierFCM, PredictionModel
from dowhy.gcm.graph import CAUSAL_MECHANISM, get_ordered_predecessors, is_root_node, validate_causal_model_assignment
from dowhy.gcm.ml import (
    create_hist_gradient_boost_classifier,
    create_hist_gradient_boost_regressor,
    create_lasso_regressor,
    create_linear_regressor,
    create_logistic_regression_classifier,
    create_random_forest_regressor,
    create_ridge_regressor,
    create_support_vector_regressor,
)
from dowhy.gcm.ml.classification import (
    create_ada_boost_classifier,
    create_extra_trees_classifier,
    create_gaussian_nb_classifier,
    create_knn_classifier,
    create_polynom_logistic_regression_classifier,
    create_random_forest_classifier,
    create_support_vector_classifier,
)
from dowhy.gcm.ml.regression import (
    create_ada_boost_regressor,
    create_extra_trees_regressor,
    create_knn_regressor,
    create_polynom_regressor,
)
from dowhy.gcm.stochastic_models import EmpiricalDistribution
from dowhy.gcm.util.general import (
    apply_one_hot_encoding,
    fit_one_hot_encoders,
    is_categorical,
    set_random_seed,
    shape_into_2d,
)

_LIST_OF_POTENTIAL_CLASSIFIERS_GOOD = [
    partial(create_logistic_regression_classifier, max_iter=1000),
    create_hist_gradient_boost_classifier,
]
_LIST_OF_POTENTIAL_REGRESSORS_GOOD = [
    create_linear_regressor,
    create_hist_gradient_boost_regressor,
]

_LIST_OF_POTENTIAL_CLASSIFIERS_BETTER = _LIST_OF_POTENTIAL_CLASSIFIERS_GOOD + [
    create_random_forest_classifier,
    create_extra_trees_classifier,
    create_support_vector_classifier,
    create_knn_classifier,
    create_gaussian_nb_classifier,
    create_ada_boost_classifier,
]
_LIST_OF_POTENTIAL_REGRESSORS_BETTER = _LIST_OF_POTENTIAL_REGRESSORS_GOOD + [
    create_ridge_regressor,
    create_polynom_regressor,
    partial(create_lasso_regressor, max_iter=5000),
    create_random_forest_regressor,
    create_support_vector_regressor,
    create_extra_trees_regressor,
    create_knn_regressor,
    create_ada_boost_regressor,
]


class AssignmentQuality(Enum):
    GOOD = auto()
    BETTER = auto()
    BEST = auto()


def assign_causal_mechanisms(
    causal_model: ProbabilisticCausalModel,
    based_on: pd.DataFrame,
    quality: AssignmentQuality = AssignmentQuality.GOOD,
    override_models: bool = False,
) -> None:
    """Automatically assigns appropriate causal models. If causal models are already assigned to nodes and
    override_models is set to False, this function only validates the assignments with respect to the graph structure.
    Here, the validation checks whether root nodes have StochasticModels and non-root ConditionalStochasticModels
    assigned.

    :param causal_model: The causal model to whose nodes to assign causal models.
    :param based_on: Jointly sampled data corresponding to the nodes of the given graph.
    :param quality: AssignmentQuality for the automatic model selection and model accuracy. This changes the type of
    prediction model and time spent on the selection. Options are:
        - AssignmentQuality.GOOD: Compares a linear, polynomial and gradient boost model on small test-training split
            of the data. The best performing model is then selected.
            Model selection speed: Fast
            Model training speed: Fast
            Model inference speed: Fast
            Model accuracy: Medium
        - AssignmentQuality.BETTER: Compares multiple model types and uses the one with the best performance
            averaged over multiple splits of the training data. By default, the model with the smallest root mean
            squared error is selected for regression problems and the model with the highest F1 score is selected for
            classification problems. For a list of possible models, see _LIST_OF_POTENTIAL_REGRESSORS_BETTER and
            _LIST_OF_POTENTIAL_CLASSIFIERS_BETTER, respectively.
            Model selection speed: Medium
            Model training speed: Fast
            Model inference speed: Fast
            Model accuracy: Good
        - AssignmentQuality.BEST: Uses an AutoGluon (auto ML) model with default settings defined by the AutoGluon
            wrapper. While the model selection itself is fast, the training and inference speed can be significantly
            slower than in the other options. NOTE: This requires the optional autogluon.tabular dependency.
            Model selection speed: Instant
            Model training speed: Slow
            Model inference speed: Slow-Medium
            Model accuracy: Best
        :param override_models: If set to True, existing model assignments are replaced with automatically selected
        ones. If set to False, the assigned models are only validated with respect to the graph structure.

    :return: None
    """
    for node in causal_model.graph.nodes:
        if not override_models and CAUSAL_MECHANISM in causal_model.graph.nodes[node]:
            validate_causal_model_assignment(causal_model.graph, node)
            continue

        if is_root_node(causal_model.graph, node):
            causal_model.set_causal_mechanism(node, EmpiricalDistribution())
        else:
            prediction_model = select_model(
                based_on[get_ordered_predecessors(causal_model.graph, node)].to_numpy(),
                based_on[node].to_numpy(),
                quality,
            )

            if isinstance(prediction_model, ClassificationModel):
                causal_model.set_causal_mechanism(node, ClassifierFCM(prediction_model))
            else:
                causal_model.set_causal_mechanism(node, AdditiveNoiseModel(prediction_model))


def select_model(
    X: np.ndarray, Y: np.ndarray, model_selection_quality: AssignmentQuality
) -> Union[PredictionModel, ClassificationModel]:
    if model_selection_quality == AssignmentQuality.BEST:
        try:
            from dowhy.gcm.ml.autolguon import AutoGluonClassifier, AutoGluonRegressor

            if is_categorical(Y):
                return AutoGluonClassifier()
            else:
                return AutoGluonRegressor()
        except ImportError:
            raise RuntimeError(
                "AutoGluon module not found! For the BEST auto assign quality, consider installing the "
                "optional AutoGluon dependency."
            )
    elif model_selection_quality == AssignmentQuality.GOOD:
        list_of_regressor = list(_LIST_OF_POTENTIAL_REGRESSORS_GOOD)
        list_of_classifier = list(_LIST_OF_POTENTIAL_CLASSIFIERS_GOOD)
        model_selection_splits = 2
    elif model_selection_quality == AssignmentQuality.BETTER:
        list_of_regressor = list(_LIST_OF_POTENTIAL_REGRESSORS_BETTER)
        list_of_classifier = list(_LIST_OF_POTENTIAL_CLASSIFIERS_BETTER)
        model_selection_splits = 5
    else:
        raise ValueError("Invalid model selection quality.")

    if apply_one_hot_encoding(X, fit_one_hot_encoders(X)).shape[1] <= 5:
        # Avoid too many features
        list_of_regressor += [create_polynom_regressor]
        list_of_classifier += [partial(create_polynom_logistic_regression_classifier, max_iter=1000)]

    if is_categorical(Y):
        return find_best_model(list_of_classifier, X, Y, model_selection_splits=model_selection_splits)()
    else:
        return find_best_model(list_of_regressor, X, Y, model_selection_splits=model_selection_splits)()


def has_linear_relationship(X: np.ndarray, Y: np.ndarray, max_num_samples: int = 3000) -> bool:
    X, Y = shape_into_2d(X, Y)

    target_is_categorical = is_categorical(Y)
    # Making sure there are at least 30% test samples.
    num_trainings_samples = min(max_num_samples, round(X.shape[0] * 0.7))
    num_test_samples = min(X.shape[0] - num_trainings_samples, max_num_samples)

    if target_is_categorical:
        all_classes, indices, counts = np.unique(Y, return_counts=True, return_index=True)
        for i in range(all_classes.size):
            # Making sure that there are at least 2 samples from one class (here, simply duplicate the point).
            if counts[i] == 1:
                X = np.row_stack([X, X[indices[i], :]])
                Y = np.row_stack([Y, Y[indices[i], :]])

        x_train, x_test, y_train, y_test = train_test_split(
            X, Y, train_size=num_trainings_samples, test_size=num_test_samples, stratify=Y
        )

    else:
        x_train, x_test, y_train, y_test = train_test_split(
            X, Y, train_size=num_trainings_samples, test_size=num_test_samples
        )

    one_hot_encoder = fit_one_hot_encoders(np.row_stack([x_train, x_test]))
    x_train = apply_one_hot_encoding(x_train, one_hot_encoder)
    x_test = apply_one_hot_encoding(x_test, one_hot_encoder)

    if target_is_categorical:
        linear_mdl = LogisticRegression(max_iter=1000)
        nonlinear_mdl = create_hist_gradient_boost_classifier()
        linear_mdl.fit(x_train, y_train.squeeze())
        nonlinear_mdl.fit(x_train, y_train.squeeze())

        # Compare number of correct classifications.
        return np.sum(shape_into_2d(linear_mdl.predict(x_test)) == y_test) >= np.sum(
            shape_into_2d(nonlinear_mdl.predict(x_test)) == y_test
        )
    else:
        linear_mdl = LinearRegression()
        nonlinear_mdl = create_hist_gradient_boost_regressor()
        linear_mdl.fit(x_train, y_train.squeeze())
        nonlinear_mdl.fit(x_train, y_train.squeeze())

        return np.mean((y_test - shape_into_2d(linear_mdl.predict(x_test))) ** 2) <= np.mean(
            (y_test - shape_into_2d(nonlinear_mdl.predict(x_test))) ** 2
        )


def find_best_model(
    prediction_model_factories: List[Callable[[], PredictionModel]],
    X: np.ndarray,
    Y: np.ndarray,
    metric: Optional[Callable[[np.ndarray, np.ndarray], float]] = None,
    max_samples_per_split: int = 10000,
    model_selection_splits: int = 5,
    n_jobs: Optional[int] = None,
) -> Callable[[], PredictionModel]:
    n_jobs = config.default_n_jobs if n_jobs is None else n_jobs

    X, Y = shape_into_2d(X, Y)

    is_classification_problem = isinstance(prediction_model_factories[0](), ClassificationModel)

    if metric is None:
        if is_classification_problem:
            metric = lambda y_true, y_preds: -metrics.f1_score(
                y_true, y_preds, average="macro", zero_division=0
            )  # Higher score is better
        else:
            metric = metrics.mean_squared_error

    labelBinarizer = None
    if is_classification_problem:
        labelBinarizer = MultiLabelBinarizer()
        labelBinarizer.fit(Y)

    kfolds = list(KFold(n_splits=model_selection_splits).split(range(X.shape[0])))

    def estimate_average_score(prediction_model_factory: Callable[[], PredictionModel], random_seed: int) -> float:
        set_random_seed(random_seed)

        average_result = 0

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ConvergenceWarning)
            for train_indices, test_indices in kfolds:
                model_instance = prediction_model_factory()
                model_instance.fit(X[train_indices[:max_samples_per_split]], Y[train_indices[:max_samples_per_split]])

                y_true = Y[test_indices[:max_samples_per_split]]
                y_pred = model_instance.predict(X[test_indices[:max_samples_per_split]])
                if labelBinarizer is not None:
                    y_true = labelBinarizer.transform(y_true)
                    y_pred = labelBinarizer.transform(y_pred)

                average_result += metric(y_true, y_pred)

        return average_result / model_selection_splits

    random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(prediction_model_factories))
    average_metric_scores = Parallel(n_jobs=n_jobs)(
        delayed(estimate_average_score)(prediction_model_factory, random_seed)
        for prediction_model_factory, random_seed in zip(prediction_model_factories, random_seeds)
    )

    return sorted(zip(prediction_model_factories, average_metric_scores), key=lambda x: x[1])[0][0]



import itertools

import pytest
from pytest import mark

from dowhy.causal_estimators.instrumental_variable_estimator import InstrumentalVariableEstimator

from .base import TestEstimator


@mark.usefixtures("fixed_seed")
class TestInstrumentalVariableEstimator(object):
    @mark.parametrize(
        [
            "error_tolerance",
            "Estimator",
            "num_common_causes",
            "num_instruments",
            "num_effect_modifiers",
            "num_treatments",
            "treatment_is_binary",
            "outcome_is_binary",
            "identifier_method",
        ],
        [
            (
                0.4,
                InstrumentalVariableEstimator,
                [0, 1],
                [1, 2],
                [
                    0,
                ],
                [1, 2],
                [False, True],
                [
                    False,
                ],
                "iv",
            ),
        ],
    )
    def test_average_treatment_effect(
        self,
        error_tolerance,
        Estimator,
        num_common_causes,
        num_instruments,
        num_effect_modifiers,
        num_treatments,
        treatment_is_binary,
        outcome_is_binary,
        identifier_method,
    ):
        estimator_tester = TestEstimator(error_tolerance, Estimator, identifier_method=identifier_method)
        # Not using testsuite from .base/TestEstimtor, custom code below
        args_dict = {
            "num_common_causes": num_common_causes,
            "num_instruments": num_instruments,
            "num_effect_modifiers": num_effect_modifiers,
            "num_treatments": num_treatments,
            "treatment_is_binary": treatment_is_binary,
            "outcome_is_binary": outcome_is_binary,
        }
        keys, values = zip(*args_dict.items())
        configs = [dict(zip(keys, v)) for v in itertools.product(*values)]
        for cfg in configs:
            print("\nConfig:", cfg)
            cfg["method_params"] = {}
            if cfg["num_instruments"] >= cfg["num_treatments"]:
                estimator_tester.average_treatment_effect_test(**cfg)
            else:
                with pytest.raises(ValueError):
                    estimator_tester.average_treatment_effect_test(**cfg)

        # More cases where Exception  is expected
        cfg = configs[0]
        cfg["num_instruments"] = 0
        with pytest.raises(ValueError):
            estimator_tester.average_treatment_effect_test(**cfg)



import numpy as np
from econml.sklearn_extensions.model_selection import GridSearchCVList
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor
from sklearn.linear_model import Lasso
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler


def get_numeric_features(X):
    """
    Finds the numeric feature columns in a dataset

    :param X: pandas dataframe

    returns: list of indices of numeric features
    """
    numeric_features_names = list(X.select_dtypes("number"))
    numeric_features = []
    for col_name in numeric_features_names:
        col_index = X.columns.get_loc(col_name)
        numeric_features.append(col_index)
    return numeric_features


def get_generic_regressor(
    cv, X, Y, max_degree=3, estimator_list=None, estimator_param_list=None, numeric_features=None
):
    """
    Finds the best estimator for regression function (g_s)

    :param cv: training and testing data indices obtained afteer Kfolding the dataset
    :param X: regressors data for training the regression model
    :param Y: outcome data for training the regression model
    :param max_degree: degree of the polynomial function used to approximate the regression function
    :param estimator_list: list of estimator objects for finding the regression function
    :param estimator_param_list: list of dictionaries with parameters for tuning respective estimators in estimator_list
    :param numeric_features: list of indices of numeric features in the dataset

    :returns: estimator for Reisz Regression function
    """
    if estimator_list is not None:
        estimator = GridSearchCVList(
            estimator_list, estimator_param_list, cv=cv, scoring="neg_mean_squared_error", n_jobs=-1
        ).fit(X, Y)
        return estimator.best_estimator_
    else:
        estimator = GridSearchCVList(
            [
                RandomForestRegressor(n_estimators=100, random_state=120),
                Pipeline(
                    [
                        (
                            "scale",
                            ColumnTransformer([("num", StandardScaler(), numeric_features)], remainder="passthrough"),
                        ),
                        ("lasso_model", Lasso()),
                    ]
                ),
                GradientBoostingRegressor(),
            ],
            param_grid_list=[
                {"n_estimators": [50], "max_depth": [3, 4, 5], "min_samples_leaf": [10, 50]},
                {"lasso_model__alpha": [0.01, 0.001, 1e-4, 1e-5, 1e-6]},
                {"learning_rate": [0.01, 0.001], "n_estimators": [50, 200]},
            ],
            cv=cv,
            scoring="neg_mean_squared_error",
            n_jobs=-1,
        ).fit(X, Y)
        return estimator.best_estimator_


def generate_moment_function(W, g):
    """
    Generate and returns moment function
    m(W,g) = g(1,W) - g(0,W) for Average Causal Effect
    """
    shape = (W.shape[0], 1)
    ones = np.ones(shape)
    zeros = np.zeros(shape)
    non_treatment_data = W[:, 1:]  # assume that treatment is one-dimensional.
    data_0 = np.hstack([zeros, non_treatment_data])  # data with treatment = 1
    data_1 = np.hstack([ones, non_treatment_data])  # data with treatment = 0
    return g(data_1) - g(data_0)


def create_polynomial_function(max_degree):
    """
    Creates a list of polynomial functions

    :param max_degree: degree of the polynomial function to be created

    :returns: list of lambda functions
    """
    polynomial_function = []
    for degree in range(max_degree + 1):

        def poly_term(x):
            return x[:, [0]] ** degree

        polynomial_function.append(poly_term)
    return polynomial_function



"""This module provides functionality for shapley value estimation.

Classes and functions in this module should be considered experimental, meaning there might be breaking API changes in
the future.
"""

import itertools
from enum import Enum
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import scipy
from joblib import Parallel, delayed
from scipy.special import comb
from sklearn.linear_model import LinearRegression
from tqdm import tqdm

import dowhy.gcm.config as config
from dowhy.gcm.constant import EPS
from dowhy.gcm.util.general import set_random_seed


class ShapleyApproximationMethods(Enum):
    """
    AUTO: Using EXACT when number of players is below 6 and EARLY_STOPPING otherwise.
    EXACT: Generate all possible subsets and estimate Shapley values with corresponding subset weights.
    EXACT_FAST: Generate all possible subsets and estimate Shapley values via weighed least squares regression. This can
                be faster, but, depending on the set function, numerically less stable.
    SUBSET_SAMPLING: Randomly samples subsets and estimate Shapley values via weighed least squares regression. Here,
                     only a certain number of randomly drawn subsets are used.
    EARLY_STOPPING: Estimate Shapley values based on a few randomly generated permutations. Stop the estimation process
                    when the the Shapley values do not change much on average anymore between runs.
    PERMUTATION: Estimates Shapley values based on a fixed number of randomly generated permutations. By fine tuning
                 hyperparameters, this can be potentially faster than the early stopping approach due to a better
                 utilization of the parallelization.
    """

    AUTO = (0,)
    EXACT = (1,)
    EXACT_FAST = (2,)
    EARLY_STOPPING = (3,)
    PERMUTATION = (4,)
    SUBSET_SAMPLING = 5


class ShapleyConfig:
    def __init__(
        self,
        approximation_method: ShapleyApproximationMethods = ShapleyApproximationMethods.AUTO,
        num_samples: int = 5000,
        min_percentage_change_threshold: float = 0.01,
        n_jobs: Optional[int] = None,
    ) -> None:
        """Config for estimating Shapley values.

        :param approximation_method: Type of approximation methods (see :py:class:`ShapleyApproximationMethods <dowhy.gcm.shapley.ShapleyApproximationMethods>`).
        :param num_samples: Number of samples used for approximating the Shapley values. Depending on the approximation
                            method, this can either represent the number of drawn subsets (in SUBSET_SAMPLING) or the
                            number of drawn permutations (in EARLY_STOPPING and PERMUTATION). In case of EARLY_STOPPING,
                            this also represents a limit on the evaluation runs.
        :param min_percentage_change_threshold: This parameter is only relevant for EARLY_STOPPING and indicates the
                                                minimum required change of the Shapley values between two runs
                                                (i.e. evaluation of permutations) before the estimation stops.
                                                For instance, if the Shapley value changes less than the given value for
                                                a certain number of consecutive runs, the algorithm stops and returns
                                                the current result.
        :param n_jobs: Number of parallel jobs.
        """
        self.approximation_method = approximation_method
        self.num_samples = num_samples
        self.min_percentage_change_threshold = min_percentage_change_threshold
        self.n_jobs = config.default_n_jobs if n_jobs is None else n_jobs


def estimate_shapley_values(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]],
    num_players: int,
    shapley_config: Optional[ShapleyConfig] = None,
) -> np.ndarray:
    """Estimates the Shapley values based on the provided set function. A set function here is defined by taking a
    (subset) of players and returning a certain utility value. This is in the context of attributing the
    value of the i-th player to a subset of players S by evaluating v(S u {i}) - v(S), where v is the
    set function and i is not in S. While we use the term 'player' here, this is often a certain feature/variable.

    The input of the set function is a binary vector indicating which player is part of the set. For instance, given 4
    players (1,2,3,4) and a subset only contains players 1,2,4, then this is indicated by the vector [1, 1, 0, 1]. The
    function is expected to return a numeric value based on this input.

    Note: The set function can be arbitrary and can resemble computationally complex operations. Keep in mind
    that the estimation of Shapley values can become computationally expensive and requires a lot of memory. If the
    runtime is too slow, consider changing the default config.

    :param set_func: A set function that expects a binary vector as input which specifies which player is part of the
                     subset.
    :param num_players: Total number of players.
    :param shapley_config: A config object for indicating the approximation method and other parameters. If None is
                           given, a default config is used. For faster runtime or more accurate results, consider
                           creating a custom config.
    :return: A numpy array representing the Shapley values for each player, i.e. there are as many Shapley values as
             num_players. The i-th entry belongs to the i-th player. Here, the set function defines which index belongs
             to which player and is responsible to keep it consistent.
    """
    if shapley_config is None:
        shapley_config = ShapleyConfig()

    approximation_method = shapley_config.approximation_method
    if approximation_method == ShapleyApproximationMethods.AUTO:
        if num_players <= 5:
            approximation_method = ShapleyApproximationMethods.EXACT
        else:
            approximation_method = ShapleyApproximationMethods.EARLY_STOPPING

    if approximation_method == ShapleyApproximationMethods.EXACT:
        return _estimate_shapley_values_exact(set_func=set_func, num_players=num_players, n_jobs=shapley_config.n_jobs)
    elif approximation_method == ShapleyApproximationMethods.PERMUTATION:
        return _approximate_shapley_values_via_permutation_sampling(
            set_func=set_func,
            num_players=num_players,
            num_permutations=max(1, shapley_config.num_samples // num_players),
            n_jobs=shapley_config.n_jobs,
        )
    elif approximation_method == ShapleyApproximationMethods.EARLY_STOPPING:
        return _approximate_shapley_values_via_early_stopping(
            set_func=set_func,
            num_players=num_players,
            max_runs=shapley_config.num_samples,
            min_percentage_change_threshold=shapley_config.min_percentage_change_threshold,
            n_jobs=shapley_config.n_jobs,
        )
    elif approximation_method == ShapleyApproximationMethods.SUBSET_SAMPLING:
        return _approximate_shapley_values_via_least_squares_regression(
            set_func=set_func,
            num_players=num_players,
            use_subset_approximation=True,
            num_samples_for_approximation=shapley_config.num_samples,
            n_jobs=shapley_config.n_jobs,
        )
    elif approximation_method == ShapleyApproximationMethods.EXACT_FAST:
        return _approximate_shapley_values_via_least_squares_regression(
            set_func=set_func,
            num_players=num_players,
            use_subset_approximation=False,
            num_samples_for_approximation=shapley_config.num_samples,
            n_jobs=shapley_config.n_jobs,
        )
    else:
        raise ValueError("Unknown method for Shapley approximation!")


def _estimate_shapley_values_exact(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]], num_players: int, n_jobs: int
) -> np.ndarray:
    """Following Eq. (2) in
    Janzing, D., Minorics, L., & Bloebaum, P. (2020).
    Feature relevance quantification in explainable AI: A causal problem.
    In International Conference on Artificial Intelligence and Statistics (pp. 2907-2916). PMLR."""
    all_subsets = [tuple(subset) for subset in itertools.product([0, 1], repeat=num_players)]

    with Parallel(n_jobs=n_jobs) as parallel:
        subset_to_result_map = _evaluate_set_function(set_func, all_subsets, parallel)

    def compute_subset_weight(length: int) -> float:
        return 1 / (num_players * comb(num_players - 1, length))

    subset_weight_cache = {}

    shapley_values = [None] * num_players
    subsets_missing_one_player = np.array(list(itertools.product([0, 1], repeat=num_players - 1)))
    for player_index in range(num_players):
        subsets_with_player = [
            tuple(subset) for subset in np.insert(subsets_missing_one_player, player_index, 1, axis=1)
        ]
        subsets_without_player = [
            tuple(subset) for subset in np.insert(subsets_missing_one_player, player_index, 0, axis=1)
        ]

        for i in range(len(subsets_with_player)):
            subset_length = int(np.sum(subsets_without_player[i]))
            if subset_length not in subset_weight_cache:
                subset_weight_cache[subset_length] = compute_subset_weight(subset_length)

            weighted_diff = subset_weight_cache[subset_length] * (
                subset_to_result_map[subsets_with_player[i]] - subset_to_result_map[subsets_without_player[i]]
            )
            # For estimating Shapley values for multiple samples (e.g. in feature relevance) and the number of samples
            # is unknown beforehand.
            if shapley_values[player_index] is None:
                shapley_values[player_index] = weighted_diff
            else:
                shapley_values[player_index] += weighted_diff

    return np.array(shapley_values).T


def _approximate_shapley_values_via_least_squares_regression(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]],
    num_players: int,
    use_subset_approximation: bool,
    num_samples_for_approximation: int,
    n_jobs: int,
    full_and_empty_subset_weight: float = 10**20,
) -> np.ndarray:
    """For more details about this approximation, see Section 4.1.1 in
    Janzing, D., Minorics, L., & Bloebaum, P. (2020).
    Feature relevance quantification in explainable AI: A causal problem.
    In International Conference on Artificial Intelligence and Statistics (pp. 2907-2916). PMLR."""
    if not use_subset_approximation:
        all_subsets, weights = _create_subsets_and_weights_exact(num_players, full_and_empty_subset_weight)
    else:
        all_subsets, weights = _create_subsets_and_weights_approximation(
            num_players, full_and_empty_subset_weight, num_samples_for_approximation
        )

    def parallel_job(subset: np.ndarray, parallel_random_seed: int):
        set_random_seed(parallel_random_seed)

        return set_func(subset)

    with Parallel(n_jobs=n_jobs) as parallel:
        random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(all_subsets))
        set_function_results = parallel(
            delayed(parallel_job)(subset, random_seed)
            for subset, random_seed in tqdm(
                zip(all_subsets, random_seeds),
                desc="Estimate shapley values as least squares solution",
                position=0,
                leave=True,
                disable=not config.show_progress_bars,
            )
        )

    return LinearRegression().fit(all_subsets, np.array(set_function_results), sample_weight=weights).coef_


def _approximate_shapley_values_via_permutation_sampling(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]], num_players: int, num_permutations: int, n_jobs: int
) -> np.ndarray:
    """For more details about this approximation, see
    Strumbelj, E., Kononenko, I. (2014).
    Explaining prediction models and individual predictions with feature contributions.
    In Knowledge and information systems, 41(3):647–665"""
    full_subset_result, empty_subset_result = _estimate_full_and_emtpy_subset_results(set_func, num_players)

    subsets_to_evaluate = set()
    all_permutations = []
    for i in range(num_permutations):
        permutation = np.random.choice(num_players, num_players, replace=False)
        all_permutations.append(permutation)

        subsets_to_evaluate.update(_create_index_order_and_subset_tuples(permutation))

    with Parallel(n_jobs=n_jobs) as parallel:
        evaluated_subsets = _evaluate_set_function(set_func, subsets_to_evaluate, parallel)

    shapley_values = _estimate_shapley_values_of_permutation(
        all_permutations[0], evaluated_subsets, full_subset_result, empty_subset_result
    )
    for i in range(1, len(all_permutations)):
        shapley_values += _estimate_shapley_values_of_permutation(
            all_permutations[i], evaluated_subsets, full_subset_result, empty_subset_result
        )

    return shapley_values / len(all_permutations)


def _approximate_shapley_values_via_early_stopping(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]],
    num_players: int,
    max_runs: int,
    min_percentage_change_threshold: float,
    n_jobs: int,
    num_permutations_per_run: int = 5,
) -> np.ndarray:
    """Combines the approximation method described in

    Strumbelj, E., Kononenko, I. (2014).
    Explaining prediction models and individual predictions with feature contributions.
    In Knowledge and information systems, 41(3):647–665

    with an early stopping criteria. This is, if the Shapley values change less than a certain threshold on average
    between two runs, then stop the estimation.
    """
    full_subset_result, empty_subset_result = _estimate_full_and_emtpy_subset_results(set_func, num_players)

    shapley_values = None
    old_shap_proxy = np.zeros(num_players)
    evaluated_subsets = {}
    num_generated_permutations = 0
    run_counter = 0
    converged_run = 0

    if config.show_progress_bars:
        pbar = tqdm(total=1)

    with Parallel(n_jobs=n_jobs) as parallel:
        # The method stops if either the change between some consecutive runs is below the given threshold or the
        # maximum number of runs is reached.
        while True:
            run_counter += 1
            subsets_to_evaluate = set()

            # In each run, we create one random permutation of players. For instance, given 4 players, a permutation
            # could be [3,1,4,2].
            permutations = [
                np.random.choice(num_players, num_players, replace=False) for _ in range(num_permutations_per_run)
            ]
            for permutation in permutations:
                num_generated_permutations += 1
                # Create all subsets belonging to the generated permutation. This is, if we have [3,1,4,2], then the
                # subsets are [3], [3,1], [3,1,4] [3,1,4,2].
                subsets_to_evaluate.update(
                    [
                        subset_tuple
                        for subset_tuple in _create_index_order_and_subset_tuples(permutation)
                        if subset_tuple not in evaluated_subsets
                    ]
                )

            # The result for each subset is cached such that if a subset that has already been evaluated appears again,
            # we can take this result directly.
            evaluated_subsets.update(_evaluate_set_function(set_func, subsets_to_evaluate, parallel, False))

            for permutation in permutations:
                # To improve the runtime, multiple permutations are evaluated in each run.
                if shapley_values is None:
                    shapley_values = _estimate_shapley_values_of_permutation(
                        permutation, evaluated_subsets, full_subset_result, empty_subset_result
                    )
                else:
                    shapley_values += _estimate_shapley_values_of_permutation(
                        permutation, evaluated_subsets, full_subset_result, empty_subset_result
                    )

            if run_counter > max_runs:
                break

            new_shap_proxy = np.array(shapley_values)
            new_shap_proxy[new_shap_proxy == 0] = EPS
            # The current Shapley values are the average of the estimated values, i.e. we need to divide by the number
            # of generated permutations here.
            new_shap_proxy /= num_generated_permutations

            if run_counter > 1:
                percentage_changes = 1 - new_shap_proxy / old_shap_proxy
                if config.show_progress_bars:
                    pbar.set_description(
                        f"Estimating Shapley Values. "
                        f"Average change of Shapley values in run {run_counter} "
                        f"({num_generated_permutations} evaluated permutations): "
                        f"{np.mean(percentage_changes) * 100}%"
                    )

                if np.mean(percentage_changes) < min_percentage_change_threshold:
                    # Here, the change between two runs is below the minimum threshold, but to reduce the likelihood
                    # that this just happened by chance, we require that this happens at least for two runs in a row.
                    converged_run += 1
                    if converged_run >= 2:
                        break
                else:
                    converged_run = 0

            old_shap_proxy = new_shap_proxy

    if config.show_progress_bars:
        pbar.update(1)
        pbar.close()

    return shapley_values / num_generated_permutations


def _create_subsets_and_weights_exact(num_players: int, high_weight: float) -> Tuple[np.ndarray, np.ndarray]:
    """Creates all subsets and the exact weights of each subset. See Section 4.1.1. in

    Janzing, D., Minorics, L., & Bloebaum, P. (2020).
    Feature relevance quantification in explainable AI: A causal problem.
    In International Conference on Artificial Intelligence and Statistics (pp. 2907-2916). PMLR.

    for more details on this.

    :param num_players: Total number of players.
    :param high_weight: A 'high' weight for computational purposes. This is used to resemble 'infinity', but needs to be
                        selected carefully to avoid numerical issues.
    :return: A tuple, where the first entry is a numpy array with all subsets and the second entry is an array with the
             corresponding weights to each subset.
    """
    all_subsets = []

    num_iterations = int(np.ceil(num_players / 2))

    for i in range(num_iterations):
        # Create all (unique) subsets)
        all_subsets.extend(
            np.array(
                [np.bincount(combs, minlength=num_players) for combs in itertools.combinations(range(num_players), i)]
            )
        )

        all_subsets.extend(
            np.array(
                [
                    np.bincount(combs, minlength=num_players)
                    for combs in itertools.combinations(range(num_players), num_players - i)
                ]
            )
        )

        if i == num_iterations - 1 and num_players % 2 == 0:
            all_subsets.extend(
                np.array(
                    [
                        np.bincount(combs, minlength=num_players)
                        for combs in itertools.combinations(range(num_players), i + 1)
                    ]
                )
            )

    weights = np.zeros(len(all_subsets))

    for i, subset in enumerate(all_subsets):
        subset_size = np.sum(subset)
        if subset_size == num_players or subset_size == 0:
            # Assigning a 'high' weight, since this resembles "infinity".
            weights[i] = high_weight
        else:
            # The weight for a subset with a specific length (see paper mentioned in the docstring for more
            # information).
            weights[i] = (num_players - 1) / (
                scipy.special.binom(num_players, subset_size) * subset_size * (num_players - subset_size)
            )

    return np.array(all_subsets, dtype=np.int), weights.astype(np.float)


def _create_subsets_and_weights_approximation(
    num_players: int, high_weight: float, num_subset_samples: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Randomly samples subsets and weights them based on the number of how often they appear.

    :param num_players: Total number of players.
    :param high_weight: A 'high' weight for computational purposes. This is used to resemble 'infinity', but needs to be
                        selected carefully to avoid numerical issues.
    :param num_subset_samples: Number of subset samples.
    :return: A tuple, where the first entry is a numpy array with the sampled subsets and the second entry is an array
             with the corresponding weights to each subset.
    """
    all_subsets = [np.zeros(num_players), np.ones(num_players)]
    weights = {tuple(all_subsets[0]): high_weight, tuple(all_subsets[1]): high_weight}

    probabilities_of_subset_length = np.zeros(num_players + 1)
    for i in range(1, num_players):
        probabilities_of_subset_length[i] = (num_players - 1) / (i * (num_players - i))

    probabilities_of_subset_length = probabilities_of_subset_length / np.sum(probabilities_of_subset_length)

    for i in range(num_subset_samples):
        subset_as_tuple = _convert_list_of_indices_to_binary_vector_as_tuple(
            np.random.choice(
                num_players, np.random.choice(num_players + 1, 1, p=probabilities_of_subset_length), replace=False
            ),
            num_players,
        )

        if subset_as_tuple not in weights:
            weights[subset_as_tuple] = 0
            all_subsets.append(np.array(subset_as_tuple))

        weights[subset_as_tuple] += 1

    weights = np.array([weights[tuple(x)] for x in all_subsets])

    return np.array(all_subsets, dtype=np.int), weights.astype(np.float)


def _convert_list_of_indices_to_binary_vector_as_tuple(list_of_indices: List[int], num_players: int) -> Tuple[int]:
    subset = np.zeros(num_players, dtype=np.int)
    subset[list_of_indices] = 1

    return tuple(subset)


def _evaluate_set_function(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]],
    evaluation_subsets: Union[Set[Tuple[int]], List[Tuple[int]]],
    parallel_context: Parallel,
    show_progressbar: bool = True,
) -> Dict[Tuple[int], Union[float, np.ndarray]]:
    def parallel_job(input_subset: Tuple[int], parallel_random_seed: int) -> Union[float, np.ndarray]:
        set_random_seed(parallel_random_seed)

        return set_func(np.array(input_subset))

    random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(evaluation_subsets))
    subset_results = parallel_context(
        delayed(parallel_job)(subset_to_evaluate, random_seed)
        for subset_to_evaluate, random_seed in tqdm(
            zip(evaluation_subsets, random_seeds),
            desc="Evaluate set function",
            position=0,
            leave=True,
            disable=not config.show_progress_bars or not show_progressbar,
        )
    )

    subset_to_result_map = {}
    for (subset, result) in zip(evaluation_subsets, subset_results):
        subset_to_result_map[subset] = result

    return subset_to_result_map


def _estimate_full_and_emtpy_subset_results(
    set_func: Callable[[np.ndarray], Union[float, np.ndarray]], num_players: int
) -> Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
    return set_func(np.ones(num_players, dtype=np.int)), set_func(np.zeros(num_players, dtype=np.int))


def _create_index_order_and_subset_tuples(permutation: List[int]) -> List[Tuple[int]]:
    indices = []
    index_tuples = []

    for var in range(len(permutation) - 1):
        indices += [permutation[var]]
        index_tuples.append(_convert_list_of_indices_to_binary_vector_as_tuple(indices, len(permutation)))

    return index_tuples


def _estimate_shapley_values_of_permutation(
    permutation: List[int],
    evaluated_subsets: Dict[Tuple[int], Union[float, np.ndarray]],
    full_subset_result: Union[float, np.ndarray],
    empty_subset_result: Union[float, np.ndarray],
) -> np.ndarray:
    current_variable_set = []
    shapley_values = [[]] * len(permutation)
    previous_result = empty_subset_result
    for n in range(len(permutation) - 1):
        current_variable_set += [permutation[n]]
        current_result = evaluated_subsets[
            _convert_list_of_indices_to_binary_vector_as_tuple(current_variable_set, len(permutation))
        ]

        shapley_values[permutation[n]] = current_result - previous_result
        previous_result = current_result

    shapley_values[permutation[-1]] = full_subset_result - previous_result

    return np.array(shapley_values).T



from abc import ABC, abstractmethod

import numpy as np


class AnomalyScorer(ABC):
    @abstractmethod
    def fit(self, X: np.ndarray) -> None:
        """Fits the anomaly scorer to the given data. Depending on the definition of the scorer, this can imply
        different things, such as fitting a (parametric) distribution to the data or estimating certain properties
        such as mean, variance, median etc. that are used for computing a score.

        :param X: Samples from the underlying data distribution.
        """
        raise NotImplementedError

    @abstractmethod
    def score(self, X: np.ndarray) -> np.ndarray:
        raise NotImplementedError



import numpy as np
import pandas as pd

from dowhy.utils.dgp import DataGeneratingProcess


class CubicDataGeneratingProcess(DataGeneratingProcess):
    """
    Implements a data generating process that returns data having cubic relationship between the treatment, outcome and confounders
    """

    POWER = 3
    NAME = "Cubic DGP"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.auto_gen = False

        if self.weights == {} and self.bias == {}:
            self.auto_gen = True

    def generate_data(self, sample_size):
        self.weights = {}
        self.bias = {}
        weights = []
        bias = []
        treatment = []
        outcome = []
        y_treatment = []
        y_control = []
        confounder = np.random.randn(sample_size, len(self.confounder))
        effect_modifier = np.random.randn(sample_size, len(self.effect_modifier))
        control_value = np.zeros((sample_size, len(self.treatment)))
        treatment_value = np.ones((sample_size, len(self.treatment)))

        for i in range(CubicDataGeneratingProcess.POWER):
            if self.auto_gen:
                self.generation_process()

            treatment.append(
                np.matmul(confounder, self.weights["confounder=>treatment"])
                + np.random.randn(sample_size, len(self.treatment))
                + self.bias["confounder=>treatment"]
            )
            outcome.append(
                np.matmul(confounder, self.weights["confounder=>outcome"])
                + np.matmul(effect_modifier, self.weights["effect_modifier=>outcome"])
                + np.matmul(treatment[i], self.weights["treatment=>outcome"])
                + self.bias["confounder=>outcome"]
            )
            y_control.append(
                np.matmul(confounder, self.weights["confounder=>outcome"])
                + np.matmul(effect_modifier, self.weights["effect_modifier=>outcome"])
                + np.matmul(control_value, self.weights["treatment=>outcome"])
                + self.bias["confounder=>outcome"]
            )
            y_treatment.append(
                np.matmul(confounder, self.weights["confounder=>outcome"])
                + np.matmul(effect_modifier, self.weights["effect_modifier=>outcome"])
                + np.matmul(treatment_value, self.weights["treatment=>outcome"])
                + self.bias["confounder=>outcome"]
            )

            weights.append(self.weights)
            bias.append(self.bias)

        treatment = treatment[0] * treatment[1] * treatment[2]
        if self.treatment_is_binary:
            treatment = self.convert_to_binary(treatment)

        outcome = outcome[0] * outcome[1] * outcome[2]
        y_control = y_control[0] * y_control[1] * y_control[2]
        y_treatment = y_treatment[0] * y_treatment[1] * y_treatment[2]
        self.true_value = np.mean(y_treatment - y_control, axis=0)

        self.weights = weights
        self.bias = bias

        return pd.DataFrame(
            np.hstack((effect_modifier, confounder, treatment, outcome)),
            columns=self.effect_modifier + self.confounder + self.treatment + self.outcome,
        )

    def generation_process(self):
        self.weights["confounder=>treatment"] = self.generate_weights((len(self.confounder), len(self.treatment)))
        self.weights["confounder=>outcome"] = self.generate_weights((len(self.confounder), len(self.outcome)))
        self.weights["effect_modifier=>outcome"] = self.generate_weights((len(self.effect_modifier), len(self.outcome)))
        self.weights["treatment=>outcome"] = self.generate_weights((len(self.treatment), len(self.outcome)))

        self.bias["confounder=>treatment"] = self.generate_bias(len(self.treatment))
        self.bias["confounder=>outcome"] = self.generate_bias(len(self.outcome))

    def generate_weights(self, dimensions):
        return np.random.randn(*dimensions)

    def generate_bias(self, dimensions):
        return np.random.randn(dimensions)

    def __str__(self):
        rep = super().__str__()

        header = """
        Cubic Data Generating Process
        ------------------------------
        """

        rep = header + rep

        return rep



"""This module defines multiple implementations of the abstract class :class:`~dowhy.gcm.graph.StochasticModel`.

Classes in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

import warnings
from typing import Dict, Optional, Tuple, Union

import numpy as np
import scipy
from scipy.stats import norm, rv_continuous, rv_discrete
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from sklearn.mixture import BayesianGaussianMixture

from dowhy.gcm.divergence import estimate_kl_divergence_continuous
from dowhy.gcm.graph import StochasticModel
from dowhy.gcm.util.general import shape_into_2d

_CONTINUOUS_DISTRIBUTIONS = [
    scipy.stats.norm,
    scipy.stats.laplace,
    scipy.stats.t,
    scipy.stats.uniform,
    scipy.stats.rayleigh,
]
_CONTINUOUS_DISTRIBUTIONS.extend(
    [
        getattr(scipy.stats, d)
        for d in dir(scipy.stats)
        if isinstance(getattr(scipy.stats, d), scipy.stats.rv_continuous) and d not in _CONTINUOUS_DISTRIBUTIONS
    ]
)

_DISCRETE_DISTRIBUTIONS = [
    getattr(scipy.stats, d) for d in dir(scipy.stats) if isinstance(getattr(scipy.stats, d), scipy.stats.rv_discrete)
]

_CONTINUOUS_DISTRIBUTIONS = {x.name: x for x in _CONTINUOUS_DISTRIBUTIONS}
_DISCRETE_DISTRIBUTIONS = {x.name: x for x in _DISCRETE_DISTRIBUTIONS}


class ScipyDistribution(StochasticModel):
    """Represents any parametric distribution that can be modeled by scipy."""

    def __init__(self, scipy_distribution: Optional[Union[rv_continuous, rv_discrete]] = None, **parameters) -> None:
        """Initializes a stochastic model that allows to sample from a parametric distribution implemented in Scipy.

        For instance, to use a beta distribution with parameters a=2 and b=0.5:
            ScipyDistribution(stats.beta, a=2, b=0.5)
        Or a Gaussian distribution with mean=0 and standard deviation 2:
            ScipyDistribution(stats.norm, loc=2, scale=0.5)

        Note that the parameter names need to coincide with the parameter names in the corresponding Scipy
        implementations. See https://docs.scipy.org/doc/scipy/tutorial/stats.html for more information.

        :param scipy_distribution: A continuous or discrete distribution parametric distribution implemented in Scipy.
        :param parameters: Set of parameters of the parametric distribution.
        """
        self._distribution = scipy_distribution
        self._parameters = parameters
        self._fixed_parameters = len(parameters) > 0

    def draw_samples(self, num_samples: int) -> np.ndarray:
        if len(self._parameters) == 0 or self._distribution is None:
            raise ValueError("Cannot draw samples. Model has not been fit!")

        return shape_into_2d(self._distribution.rvs(size=num_samples, **self.parameters))

    def fit(self, X: np.ndarray) -> None:
        if self._distribution is None:
            # Currently only support continuous distributions for auto selection.
            best_model, best_parameters = self.find_suitable_continuous_distribution(X)
            self._distribution = best_model
            self._parameters = best_parameters
        elif not self._fixed_parameters:
            self._parameters = self.map_scipy_distribution_parameters_to_names(
                self._distribution, self._distribution.fit(shape_into_2d(X))
            )

    @property
    def parameters(self) -> Dict[str, float]:
        return self._parameters

    @property
    def scipy_distribution(self) -> Optional[Union[rv_continuous, rv_discrete]]:
        return self._distribution

    def clone(self):
        if self._fixed_parameters:
            return ScipyDistribution(scipy_distribution=self._distribution, **self._parameters)
        else:
            return ScipyDistribution(scipy_distribution=self._distribution)

    @staticmethod
    def find_suitable_continuous_distribution(
        distribution_samples: np.ndarray, divergence_threshold: float = 10**-2
    ) -> Tuple[rv_continuous, Dict[str, float]]:
        """Tries to find the best fitting continuous parametric distribution of given samples. This is done by fitting
        different parametric models and selecting the one with the smallest KL divergence between observed and generated
        samples.
        """
        distribution_samples = shape_into_2d(distribution_samples)

        currently_best_distribution = norm
        currently_best_parameters = (0.0, 1.0)
        currently_smallest_divergence = np.inf

        # Estimate distribution parameters from data.
        for distribution in _CONTINUOUS_DISTRIBUTIONS.values():
            # Ignore warnings from fitting process.
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")

                try:
                    # Fit distribution to data.
                    params = distribution.fit(distribution_samples)
                except ValueError:
                    # Some distributions might not be compatible with the data.
                    continue

                # Separate parts of parameters.
                arg = params[:-2]
                loc = params[-2]
                scale = params[-1]

                generated_samples = distribution.rvs(size=distribution_samples.shape[0], loc=loc, scale=scale, *arg)

                # Check the KL divergence between the distribution of the given and fitted distribution.
                divergence = estimate_kl_divergence_continuous(distribution_samples, generated_samples)
                if divergence < divergence_threshold:
                    currently_best_distribution = distribution
                    currently_best_parameters = params
                    break

                # Identify if this distribution is better.
                if currently_smallest_divergence > divergence:
                    currently_best_distribution = distribution
                    currently_best_parameters = params
                    currently_smallest_divergence = divergence

        return currently_best_distribution, ScipyDistribution.map_scipy_distribution_parameters_to_names(
            currently_best_distribution, currently_best_parameters
        )

    @staticmethod
    def map_scipy_distribution_parameters_to_names(
        scipy_distribution: Union[rv_continuous, rv_discrete], parameters: Tuple[float]
    ) -> Dict[str, float]:
        """Helper function to obtain a mapping from parameter name to parameter value. Depending whether the
        distribution is discrete or continuous, there are slightly different parameter names. The given parameters are
        assumed to follow the order as provided by the scipy fit function.

        :param scipy_distribution: The scipy distribution.
        :param parameters: The values of the corresponding parameters of the distribution. Here, it is expected to
                           follow the same order as defined by the scipy fit function.
        :return: A dictionary that maps a parameter name to its value.
        """
        if scipy_distribution.shapes:
            parameter_list = [name.strip() for name in scipy_distribution.shapes.split(",")]
        else:
            parameter_list = []
        if scipy_distribution.name in _DISCRETE_DISTRIBUTIONS:
            parameter_list += ["loc"]
        elif scipy_distribution.name in _CONTINUOUS_DISTRIBUTIONS:
            parameter_list += ["loc", "scale"]
        else:
            raise ValueError(
                "Distribution %s not found in the list of continuous and discrete distributions!"
                % scipy_distribution.name
            )

        parameters_dictionary = {}
        for i, parameter_name in enumerate(parameter_list):
            parameters_dictionary[parameter_name] = parameters[i]

        return parameters_dictionary


class EmpiricalDistribution(StochasticModel):
    """An implementation of a stochastic model that uniformly samples from data samples. By randomly returning a sample
    from the training data set, this model represents a parameter free representation of the marginal distribution of
    the training data. However, it will not generate unseen data points. For this, consider :py:class:`BayesianGaussianMixtureDistribution <dowhy.gcm.BayesianGaussianMixtureDistribution>`.
    """

    def __init__(self) -> None:
        self._data = None

    @property
    def data(self) -> np.ndarray:
        return self._data

    def fit(self, X: np.ndarray) -> None:
        self._data = shape_into_2d(X)

    def draw_samples(self, num_samples: int) -> np.ndarray:
        if self.data is None:
            raise RuntimeError("%s has not been fitted!" % self.__class__.__name__)

        return self.data[np.random.choice(self.data.shape[0], size=num_samples, replace=True), :]

    def clone(self):
        return EmpiricalDistribution()


class BayesianGaussianMixtureDistribution(StochasticModel):
    def __init__(self) -> None:
        self._gmm_model = None

    def fit(self, X: np.ndarray) -> None:
        X = shape_into_2d(X)
        self._gmm_model = BayesianGaussianMixture(
            n_components=BayesianGaussianMixtureDistribution._get_optimal_number_of_components(X), max_iter=1000
        ).fit(X)

    @staticmethod
    def _get_optimal_number_of_components(X: np.ndarray) -> int:
        current_best = 0
        current_best_num_components = 1
        num_best_in_succession = 0
        try:
            for i in range(2, int(np.sqrt(X.shape[0] / 2))):
                kmeans = KMeans(n_clusters=i).fit(X)
                coefficient = silhouette_score(X, kmeans.labels_, sample_size=5000)

                if coefficient > current_best:
                    current_best = coefficient
                    current_best_num_components = i
                    num_best_in_succession = 0
                else:
                    num_best_in_succession += 1

                if num_best_in_succession >= 3:
                    break
        except ValueError:
            # This error is typically raised when the data is discrete and all points are assigned to less cluster than
            # specified. It can also happen due to duplicated points. In these cases, the current best solution should
            # be sufficient.
            return current_best_num_components

        return current_best_num_components

    def draw_samples(self, num_samples: int) -> np.ndarray:
        if self._gmm_model is None:
            raise RuntimeError("%s has not been fitted!" % self.__class__.__name__)

        return shape_into_2d(self._gmm_model.sample(num_samples)[0])

    def __str__(self) -> str:
        return "Approximated data distribution"

    def clone(self):
        return BayesianGaussianMixtureDistribution()



import networkx as nx
import numpy as np

from dowhy.graph_learners import GraphLearner
from dowhy.utils.graph_operations import *

from . import get_library_class_object


class CDT(GraphLearner):
    """
    Causal discivery using the Causal Discovery Toolbox.
    Link: https://github.com/FenTechSolutions/CausalDiscoveryToolbox
    """

    def __init__(self, data, full_method_name, *args, **kwargs):
        super().__init__(data, full_method_name, *args, **kwargs)

        library_class = get_library_class_object(full_method_name)
        self._method = library_class(*args, **kwargs)

    def learn_graph(self, labels=None):
        """
        Discover causal graph and return the graph in DOT format.

        """
        graph = self._method.predict(self._data)

        # Get adjacency matrix
        self._adjacency_matrix = nx.to_numpy_matrix(graph)
        self._adjacency_matrix = np.asarray(self._adjacency_matrix)

        # If labels not provided
        if labels is not None:
            self._labels = labels

        self._graph_dot = adjacency_matrix_to_graph(self._adjacency_matrix, self._labels)

        # Obtain valid DOT format
        self._graph_dot = str_to_dot(self._graph_dot.source)
        return self._graph_dot



import random

import pandas as pd
import pytest
from pytest import mark

import dowhy.datasets

from .base import TestRefuter


@mark.usefixtures("fixed_seed")
class TestPlaceboRefuter(object):
    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_samples"], [(0.03, "backdoor.linear_regression", 1000)]
    )
    def test_refutation_placebo_refuter_continuous(self, error_tolerance, estimator_method, num_samples):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "placebo_treatment_refuter")
        refuter_tester.continuous_treatment_testsuite(num_samples=num_samples)  # Run both

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_samples"], [(0.1, "backdoor.propensity_score_matching", 5000)]
    )
    def test_refutation_placebo_refuter_binary(self, error_tolerance, estimator_method, num_samples):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "placebo_treatment_refuter")
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_samples"], [(0.1, "backdoor.linear_regression", 5000)]
    )
    def test_refutation_placebo_refuter_category(self, error_tolerance, estimator_method, num_samples):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "placebo_treatment_refuter")
        refuter_tester.categorical_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerance", "estimator_method", "num_samples"], [(0.1, "backdoor.linear_regression", 5000)]
    )
    def test_refutation_placebo_refuter_category_non_consecutive_index(
        self, error_tolerance, estimator_method, num_samples
    ):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "placebo_treatment_refuter")
        data = dowhy.datasets.linear_dataset(
            beta=10,
            num_common_causes=1,
            num_instruments=1,
            num_samples=num_samples,
            treatment_is_binary=False,
            treatment_is_category=True,
        )
        random_index = random.sample(range(1, 10 * num_samples), num_samples)
        data["df"].index = random_index
        refuter_tester.null_refutation_test(data=data)



import logging

import numpy as np

from dowhy.causal_refuter import CausalRefutation, CausalRefuter
from dowhy.utils.cit import conditional_MI, partial_corr


class GraphRefuter(CausalRefuter):
    """
    Class for performing refutations on graph and storing the results
    """

    def __init__(
        self, data, method_name_discrete="conditional_mutual_information", method_name_continuous="partial_correlation"
    ):
        """
        Initialize data for graph refutation

        :param data:input dataset
        :param method_name_discrete: name of method for testing conditional independence in discrete data
        :param method_name_continuous: name of method for testing conditional independece in continuous data
        :returns : instance of GraphRefutation class
        """
        self._refutation_passed = None
        self._data = data
        self._method_name_discrete = method_name_discrete
        self._method_name_continuous = method_name_continuous
        self._false_implications = []  # List containing the implications from the graph which hold false for dataset
        self._true_implications = []  # List containing the implications from the graph which hold true for dataset
        self._results = {}  # A dictionary with key as test set and value as [p-value, test_result]
        self.logger = logging.getLogger(__name__)

    def set_refutation_result(self, number_of_constraints_model):
        """
        Method to set the result for graph refutation. Set true if there are no false implications else false
        """
        if (len(self._true_implications)) == number_of_constraints_model:
            self._refutation_passed = True
        elif len(self._false_implications) == 0:
            self._refutation_passed = True
            self.logger.warn("Some tests could not be run : config not supported")
        elif len(self._false_implications) > 0:
            self._refutation_passed = False

    def partial_correlation(self, x=None, y=None, z=None):
        stats = partial_corr(data=self._data, x=x, y=y, z=list(z))
        p_value = stats["p-val"]
        key = (x, y) + (z,)
        if p_value < 0.05:
            # Reject H0
            self._false_implications.append([x, y, z])
            self._results[key] = [p_value, False]
        else:
            self._true_implications.append([x, y, z])
            self._results[key] = [p_value, True]

    def conditional_mutual_information(self, x=None, y=None, z=None):
        cmi_val = conditional_MI(data=self._data, x=x, y=y, z=list(z))
        key = (x, y) + (z,)
        if cmi_val <= 0.05:
            self._true_implications.append([x, y, z])
            self._results[key] = [cmi_val, True]
        else:
            self._false_implications.append([x, y, z])
            self._results[key] = [cmi_val, False]

    def refute_model(self, independence_constraints):
        """
        Method to test conditional independence using the graph refutation object on the given testing set

        :param independence_constraints: List of implications to test the conditional independence on
        :returns : GraphRefutation object
        """

        refute = GraphRefutation(
            method_name_continuous=self._method_name_continuous, method_name_discrete=self._method_name_discrete
        )

        all_nodes = list(self._data.columns.values)
        discrete_columns = []
        continuous_columns = []
        binary_columns = []
        variable_type = dict()
        for node in all_nodes:
            if self._data[node].dtype == np.int64 or self._data[node].dtype == np.int32:
                discrete_columns.append(node)
                variable_type[node] = "discrete"
                if self._data[node].isin([0, 1]).all():
                    binary_columns.append(node)
                    variable_type[node] = "binary"
            else:
                continuous_columns.append(node)
                variable_type[node] = "continuous"
        for a, b, c in independence_constraints:
            if a in continuous_columns and b in continuous_columns and all(node in continuous_columns for node in c):
                # a, b and c are all continuous variables
                if self._method_name_continuous is None or self._method_name_continuous == "partial_correlation":
                    self.partial_correlation(x=a, y=b, z=c)
                else:
                    self.logger.error(
                        "Invalid conditional independence test for continuous data. Supported tests - partial_correlation"
                    )

            elif a in discrete_columns and b in discrete_columns and all(node in discrete_columns for node in c):
                # a, b and c are all discrete variables
                if self._method_name_discrete is None or self._method_name_discrete == "conditional_mutual_information":
                    self.conditional_mutual_information(x=a, y=b, z=c)
                else:
                    self.logger.error(
                        "Invalid conditional independence test for discrete data. Supported tests - conditional_mutual_information"
                    )

            elif (
                (a in continuous_columns or a in binary_columns)
                and (b in continuous_columns or b in binary_columns)
                and all(node in continuous_columns or node in binary_columns for node in c)
            ):
                # c is set of continuous and binary variables and
                #   1. either a and b is continuous and the other is binary
                #   2. both a and b are binary
                self.partial_correlation(x=a, y=b, z=c)

            elif all(node in discrete_columns for node in c) and (a in discrete_columns or b in discrete_columns):
                # c is discrete and
                # either a or b is continuous and the other is discrete
                self.conditional_mutual_information(x=a, y=b, z=c)

            elif a in discrete_columns and b in discrete_columns:
                # a and b are discrete and c is a mixture of discrete and continuous variables. We discretize c and calculate conditional mutual information
                self.conditional_mutual_information(x=a, y=b, z=c)

            else:
                key = (a, b) + (c,)
                self._results[key] = [None, "NotImplemented"]
                variable_types_c = []
                for var in c:
                    variable_types_c.append(variable_type[var])
                self.logger.warning(
                    "The following setting with {0} as {1}, {2} as {3}, {4} as {5} not supported".format(
                        a, variable_type[a], b, variable_type[b], c, variable_types_c
                    )
                )

        self.set_refutation_result(number_of_constraints_model=len(independence_constraints))
        refute.add_conditional_independence_test_result(
            number_of_constraints_model=len(independence_constraints),
            number_of_constraints_satisfied=len(self._true_implications),
            refutation_result=self._refutation_passed,
        )
        return refute


class GraphRefutation(CausalRefutation):
    """Class for storing the result of a refutation method."""

    def __init__(self, method_name_discrete, method_name_continuous):
        self.method_name_discrete = method_name_discrete
        self.method_name_continuous = method_name_continuous
        self.number_of_constraints_model = None
        self.number_of_constraints_satisfied = None
        self.refutation_result = None

    def add_conditional_independence_test_result(
        self, number_of_constraints_model, number_of_constraints_satisfied, refutation_result
    ):
        self.number_of_constraints_model = number_of_constraints_model
        self.number_of_constraints_satisfied = number_of_constraints_satisfied
        self.refutation_result = refutation_result

    def __str__(self):
        if self.refutation_result is None:
            return "Method name for discrete data:{0}\nMethod name for continuous data:{1}".format(
                self.method_name_discrete, self.method_name_continuous
            )
        else:
            return "Method name for discrete data:{0}\nMethod name for continuous data:{1}\nNumber of conditional independencies entailed by model:{2}\nNumber of independences satisfied by data:{3}\nTest passed:{4}\n".format(
                self.method_name_discrete,
                self.method_name_continuous,
                self.number_of_constraints_model,
                self.number_of_constraints_satisfied,
                self.refutation_result,
            )



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os

# -- Project information -----------------------------------------------------

project = "DoWhy"
copyright = "2022, PyWhy contributors"
author = "PyWhy community"
version = os.environ.get("CURRENT_VERSION")

# Version Information (for version-switcher)
not_empty = lambda x: len(x) > 0
to_tag_obj = lambda t: {"name": t, "url": f"/dowhy/{t}/index.html"}
has_doc = lambda t: os.path.exists(f"../../dowhy-docs/{t}/index.html")

git_tags = reversed(list(filter(not_empty, os.environ.get("TAGS").split(","))))
doc_tags = list(filter(has_doc, git_tags))

html_context = {
    "current_version": {"name": os.environ.get("CURRENT_VERSION")},
    "versions": {
        "tags": list(map(to_tag_obj, doc_tags)),
        "branches": [{"name": "main"}],
    },
}

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "nbsphinx",
    "sphinx_rtd_theme",
    "sphinxcontrib.googleanalytics",
    "sphinx_copybutton",
    "sphinx_design",
]

# sphinx-panels shouldn't add bootstrap css since the pydata-sphinx-theme
# already loads it
panels_add_bootstrap_css = False

googleanalytics_id = "G-B139P18WHM"
copybutton_prompt_text = ">>> "

autodoc_mock_imports = ["matplotlib", "causalml", "pymc3", "econml"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    ".ipynb_checkpoints",
    "example_notebooks/dowhy_ranking_methods.ipynb",
    "example_notebooks/dowhy_causal_discovery_example.ipynb",  # need to check why str_to_dot fails here
    "example_notebooks/dowhy_twins_example.ipynb",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"navbar_end": ["navbar-icon-links", "versions"]}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "dowhydoc"

html_logo = "_static/dowhy-logo-small.png"

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "dowhy.tex", "dowhy Documentation", "Author", "manual"),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "dowhy", "dowhy Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, "dowhy", "dowhy Documentation", author, "dowhy", "One line description of project.", "Miscellaneous"),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# init docstrings should also be included in class
autoclass_content = "both"
